package feud.clientserver;

import java.net.*;
import java.util.*;
import java.io.*;

/**
 * As the name indicates, this is a simple server class. Once started, it opens a TCP-socket for clients to connect
 * as well as an (optional) UDP datagram socket (if possible) that broadcasts this server's ip and port for allowing
 * to implement a simple server browser.<br>
 * The communication between this server and its clients is done via DataContainers, which are simple Wrappers for
 * primitive values (and Strings) into byte arrays.
 */
public class SimpleServer {

    /**
     * The default tcp port
     */
    public final static int DEFAULT_PORT = 7000;

    /**
     * The default UDP port
     */
    public final static int UDP_DEFAULT_PORT = 7001;

    private List listener = new ArrayList();
    private List logoutListener = new ArrayList();
    private List loginListener = new ArrayList();
    private Map clientThreads = new HashMap();
    private PerformanceCounter pc = new PerformanceCounter("Server");
    private boolean terminate = false;
    private String name = "Default";
    private boolean registerRunning = false;
    private boolean broadcastRunning = false;

    /**
     * Creates a new Server. When creating a two server, up to two threads are being spawned. One for
     * listening on the tcp socket for clients to connect and one for broadcasting the server's data
     * via UDP.
     * @param tcpPort The tcp port to which a client can connect with this server. This port has to be open and unused or otherwise, the server won't start.
     * @param udpPort the udp port for broadcasting the server's ip and port (and player count). If this port is not available, the broadcasting thread will
     * terminate, but the server will run anyway.
     * @param doBroadcast If true, the broadcasting thread will be spawned. Otherwise, it won't.
     * @param serverName The name of the server. Should be unique but doesn't have to.
     */
    public SimpleServer(int tcpPort, int udpPort, boolean doBroadcast, String serverName) {
        if (serverName != null) {
            name = serverName;
        }
        new Thread(new ClientRegisterService(tcpPort)).start();
        if (doBroadcast) {
            new Thread(new ServerBroadcast(tcpPort, udpPort)).start();
        }
    }

    /**
     * Adds a new DataTransferListener to the server. A DataTransferListener will be notified in case of data
     * being transfered from the client to the server.
     * @param sl the listener
     */
    public void addListener(DataTransferListener sl) {
        listener.add(sl);
    }

    /**
     * Adds a new ClientLogoutListener to the server. This listener will be notified if a client logs out
     * from this server.
     * @param ctl the listener
     */
    public void addLogoutListener(ClientLogoutListener ctl) {
        logoutListener.add(ctl);
    }

    /**
     * Adds a new ClientLoginListener to the server. This listener will be notified if a client logs in
     * on this server.
     * @param ctl the listener
     */
    public void addLoginListener(ClientLoginListener ctl) {
        loginListener.add(ctl);
    }

    /**
     * Shuts down the server. This shut down is semi-hard, i.e. all server threads and such will be terminated correctly, but
     * there's no waiting for the clients to be informed about this. The clients have to take care of a "lost" server themselves.
     */
    public void shutDown() {
        terminate = true;
        try {
            while (registerRunning || broadcastRunning) {
                Thread.sleep(20);
            }
        } catch (Exception e) {
            // Who cares...?
            e.printStackTrace();
        }
    }

    /**
     * The broadcast thread. The broadcast goes to all clients in the local subnet via UDP.
     * If the port is not available, this thread will terminate with a RuntimeException but the
     * server will continue to run without it.
     * Currently, the server will send its data into the network every three seconds.
     */
    private class ServerBroadcast implements Runnable {

        private int port = 0;
        private int tcpPort = 0;

        public ServerBroadcast(int tcpPort, int port) {
            this.port = port;
            this.tcpPort = tcpPort;
        }

        public void run() {
            try {
                DatagramSocket bsock = new DatagramSocket(port);
                InetAddress bc = InetAddress.getByName("255.255.255.255");
                broadcastRunning = true;

                while (!terminate) {
                    try {
                        NetLogger.log("Server: Broadcasting server data!");
                        DataContainer dc = new DataContainer();
                        dc.add(name);
                        dc.add(tcpPort);
                        dc.add(clientThreads.size());
                        byte[] bytes = DataContainerFactory.toByteArray(new DataContainer[] {dc}, false);
                        DatagramPacket dps = new DatagramPacket(bytes, bytes.length, bc, port + 1);
                        bsock.send(dps);
                        Thread.sleep(3000);
                    } catch (Exception e) {
                        // Don't stop...just take notice!
                        e.printStackTrace();
                    }
                }
                bsock.close();
                broadcastRunning = false;
                NetLogger.log("Server: Broadcast thread terminated!");

            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
    }


    /**
     * This is the server's listening thread. It waits for client connections on the specified tcp port and spawns a new
     * thread for each client that handles this client's requests and responses. If this thread dies (or doesn't start up
     * due to the port not being available), the server in not running.
     */
    private class ClientRegisterService implements Runnable {

        private int port = 0;

        public ClientRegisterService(int port) {
            this.port = port;
        }

        public void run() {
            try {
                ServerSocket socket = new ServerSocket(port);
                socket.setSoTimeout(5000);
                registerRunning = true;
                NetLogger.log("Server: Listening for client requests at port: " + port);
                while (!terminate) {
                    try {
                        Socket sock = socket.accept();
                        NetLogger.log("Server: Request from client " + sock.getInetAddress() + "!");
                        InputStream is = sock.getInputStream();
                        OutputStream os = sock.getOutputStream();

                        byte[] bytes = StreamConverter.convert(is);

                        DataContainer[] cs = DataContainerFactory.extractContainers(bytes, false);

                        DataContainer c = cs[0];
                        DataContainer add = null;
                        if (cs.length > 1) {
                            add = cs[1];
                        }
                        DataContainer res = new DataContainer();
                        boolean ok = false;
                        if (c.hasData()) {
                            NetLogger.log("Server: Checking request type!");

                            int command = c.getMessageType();
                            // Login
                            if (command == MessageTypes.LOGIN_REQUEST) {
                                ClientInfo ci = new ClientInfo(sock.getInetAddress(), sock.getPort());
                                if (!clientThreads.containsKey(ci)) {
                                    byte zip = c.getNextByte();
                                    NetLogger.log("Server: Request type is 'login', clientID will be " + ci.getID() + ", zip=" + (zip == 1) + "!");
                                    ci.setZipMode(zip == 1);
                                    res.setMessageType(MessageTypes.LOGIN_SUCCESS);
                                    res.add(ci.getID());

                                    ClientProcessor cp = new ClientProcessor(sock, ci);
                                    clientThreads.put(ci, cp);

                                    ok = true;

                                    for (Iterator itty = loginListener.iterator(); itty.hasNext(); ) {
                                        ClientLoginListener cll = (ClientLoginListener) itty.next();
                                        DataContainer resp = cll.loggedIn(ci, add);
                                        if (resp != null) {
                                            broadcast(resp);
                                        } else {
                                            ok = false;
                                            break;
                                        }
                                    }

                                    if (ok) {
                                        new Thread(cp).start();
                                    }
                                } else {
                                    // @todo: May this happen?
                                }
                            }
                        }
                        if (!ok) {
                            res.setMessageType(MessageTypes.LOGIN_FAILURE);
                            NetLogger.log("Server: Request type is 'unknown' or no request type found (" + c.hasData() + ")!");
                        }
                        // Send response
                        os.write(DataContainerFactory.toByteArray(new DataContainer[] {res}, false));
                        os.flush();
                    } catch (SocketTimeoutException se) {
                        // Nobody cares...
                    } catch (Exception e) {
                        // This thread must not die. But it may print out its opinions on the current situation...
                        e.printStackTrace();
                    }
                }
                socket.close();
                registerRunning = false;
                NetLogger.log("Server: Register thread terminated!");
            }

            catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
    }


    /**
     * Adds a DataContainer to the broadcast-queue. Each container added to this queue is ensured to be send
     * once to all clients known to the server at the time that this method is being called (unless the client
     * hasn't died or logout, of course).<br>
     * This has nothing to do with the data broadcast via UDP that the ServerBroadcast is doing. This is a tcp
     * "broadcast" to all known clients.
     * @param dc the DataContainer that should be enqueued
     */
    public void broadcast(DataContainer dc) {
        if (dc != null) {
            for (Iterator itty2 = clientThreads.values().iterator(); itty2.hasNext(); ) {
                ClientProcessor cpt = (ClientProcessor) itty2.next();
                cpt.hasToSend(dc);
            }
        }
    }

    /**
     * This is the worker thread. Each client connecting will cause an instance of this to be spawned. Each client's
     * thread terminates if the client logs or times out. Time out time is three seconds.
     */
    private class ClientProcessor implements Runnable {

        private InputStream is = null;
        private OutputStream os = null;
        private ClientInfo ci = null;
        private Socket sock = null;
        private boolean exit = false;
        private List hasToSend = new ArrayList();

        /**
         * Creates a instance with the given socket and the ClientInfo that identifies this client.
         * @param sock the socket to which is client is connected
         * @param ci the ClientInfo
         * @throws Exception if anything goes wrong...
         */
        public ClientProcessor(Socket sock, ClientInfo ci) throws Exception {
            sock.setSoTimeout(3000);
            this.is = sock.getInputStream();
            this.os = sock.getOutputStream();
            this.ci = ci;
            this.sock = sock;
        }

        /**
         * Stops a client's thread and removes the client from the server's clients list.
         */
        public void kill() {
            exit = true;
        }

        /**
         * Used to enqueue a DataContainer into this client's broadcast queue.
         * @param dc the container
         */
        public void hasToSend(DataContainer dc) {
            if (dc != null) {
                synchronized (hasToSend) {
                    hasToSend.add(dc);
                }
            }
        }

        public void run() {
            try {
                while (!exit && !terminate) {
                    boolean recd = false;
                    DataContainer c = null;
                    DataContainer[] cs = null;
                    try {
                        byte[] bytes = StreamConverter.convert(is);
                        cs = DataContainerFactory.extractContainers(bytes, ci.getZipMode());
                        recd = true;
                    } catch (Exception e) {
                        e.printStackTrace();
                        NetLogger.log("Server: Client lost (timeout?)!");
                        logout(ci);
                        exit = true;
                    }
                    pc.printStats();
                    if (recd) {

                        List res = new ArrayList();
                        for (int i = 0; i < cs.length; i++) {
                            c = cs[i];
                            pc.in(c.getLength());
                            c.setClientInfo(ci);
                            int msgType = c.getMessageType();

                            if (msgType <= MessageTypes.INTERNAL_BORDER) {
                                // "Normal" transmission, i.e. no internal server message
                                for (Iterator itty = listener.iterator(); itty.hasNext(); ) {
                                    DataTransferListener sl = (DataTransferListener) itty.next();
                                    DataContainer[] crs = sl.dataReceived(c, msgType);
                                    if (crs != null) {
                                        res.addAll(Arrays.asList(crs));
                                    }
                                }
                            } else {
                                // an internal server message
                                processMessage(msgType, res);
                            }
                        }

                        // Notify all listeners that data has been received...
                        for (Iterator itty = listener.iterator(); itty.hasNext(); ) {
                            DataTransferListener sl = (DataTransferListener) itty.next();
                            sl.dataReceivedEnd();
                        }

                        // Send response to the client...
                        DataContainer[] crs = null;
                        int pos = 0;
                        synchronized (hasToSend) {
                            crs = new DataContainer[res.size() + hasToSend.size()];

                            // Append all responses from the transfer listeners...
                            for (Iterator itty = res.iterator(); itty.hasNext(); ) {
                                DataContainer cr = (DataContainer) itty.next();
                                crs[pos] = cr;
                                pos++;
                            }
                            // Append all data from the broadcast queue...
                            for (Iterator itty2 = hasToSend.iterator(); itty2.hasNext(); ) {
                                DataContainer dhts = (DataContainer) itty2.next();
                                if (dhts == null) {
                                    throw new RuntimeException("Null value in broascast queue at position " + (pos - 1));
                                }
                                crs[pos] = dhts;
                                pos++;
                            }
                            hasToSend.clear();
                        }

                        byte[] bytes = DataContainerFactory.toByteArray(crs, ci.getZipMode());
                        os.write(bytes);
                        pc.out(bytes.length);

                        os.flush();
                        ci.touch();
                    }
                    // Each thread runs once every 15 seconds. This may have to be adjusted for some
                    // applications...
                    Thread.sleep(15);
                }

                // This client has logged out, died, failed...whatever...we are done with it.

                logout(ci);
                clientThreads.remove(ci);

                is.close();
                os.close();
                sock.close();
                NetLogger.log("Server: Thread for client " + ci + " terminated!");
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }

        /**
         * Processes an internal server message. Currently, this can be either log out...or log out...:-)
         * @param msgType int
         * @param res List
         */
        private void processMessage(int msgType, List res) {
            DataContainer cr = new DataContainer();
            cr.setZip(ci.getZipMode());

            if (msgType == MessageTypes.LOGOUT_REQUEST) {
                NetLogger.log("Server: Request type is 'logout' from client " + ci + "!");
                cr.setMessageType(MessageTypes.LOGOUT_SUCCESS);
                kill();
                logout(ci);
            }
            res.add(cr);
        }
    }


    /**
     * Logout a client by notifying all listeners and adding their responses to the broadcast
     * queue.
     * @param ci the client to log out
     */
    private void logout(ClientInfo ci) {
        if (clientThreads.containsKey(ci) && !ci.isLoggedOut()) {
            for (Iterator itty2 = logoutListener.iterator(); itty2.hasNext(); ) {
                ClientLogoutListener ctl = (ClientLogoutListener) itty2.next();
                DataContainer dc = ctl.loggedOut(ci);
                broadcast(dc);
                NetLogger.log("Server: Client " + ci.getAddress() + " logged out!");
            }
            ci.logout();
        }
    }
}
