package feud.testgame;

import java.awt.Color;
import java.io.*;
import java.util.*;

import com.threed.jpct.*;
import com.threed.jpct.util.*;
import feud.clientserver.*;
import feud.testgame.view.*;
import feud.testgame.model.*;
import feud.testgame.util.*;
import feud.testgame.gui.*;

/**
 * The client, i.e. the actual game. This client also opens a window with a simple server browser to join
 * an existing server or start a new one.
 */
public class BlueThunderClient implements DataTransferListener, ClientPreProcessor, GameClient {

    private final static Object SYNC = new Object();
    private final static Object KILL_SYNC = new Object();

    private final static int WIDTH = 800;
    private final static int HEIGHT = 600;

    private Level level = null;
    private World world = null;
    private FrameBuffer buffer = null;
    private KeyMapper keyMapper = null;

    private World sky = null;
    private Object3D dome = null;

    private LocalPlayerObject player = null;

    private IPostProcessor blo = null;

    private ClientObjectManager coMan = null;

    private SimpleClient clientImpl = null;
    private Ticker timeout = null;
    private KeyStates keyStates = null;
    private int videoMode = FrameBuffer.SAMPLINGMODE_HARDWARE_ONLY;

    private LocalBulletManager bulMan = null;
    private DecalManager decals = null;
    private Ticker fireTicker = null;
    private List toKill = new ArrayList();
    private ServerBrowser browser = null;
    private ServerSelection serverSel = null;
    private MouseMapper mouse = null;
    private ServerEntry selectedServer = null;
    private BlueThunderServer serverImpl = null;
    private boolean quit = false;
    private Highscores hiView = null;
    private HighscoreTable tmpScores = null;
    private boolean received = false;
    private boolean shouldBeConnected = false;

    private boolean firstTransferFinished = false;
    private boolean entitiesUpdated = false;

    private Object3D crossHair = null;

    private Table messages = null;
    private Label clickMessage = null;
    private SimpleVector spawnPoint = new SimpleVector(0, -150, 0);
    private SimpleVector[] spawnPoints = new SimpleVector[] {new SimpleVector( -132, -21, 32), new SimpleVector( -146, -21, -84),
                                         new SimpleVector( -63, -2, -48), new SimpleVector(13, -21, -73), new SimpleVector(91, 3, 133), new SimpleVector( -26, -1, 140)};


    /**
     * Create a new client. This starts the client and enters the game loop.
     * @throws Exception
     */
    public BlueThunderClient() throws Exception {
        run();
    }

    public void connect(ServerEntry server, String name) throws Exception {
        try {
            while (mouse.buttonDown(0)) {
                // Avoid entering the level while the mouse button is being pressed
                org.lwjgl.input.Mouse.poll();
                Thread.sleep(10);
            }
            player.setPosition(new SimpleVector(0, -10000, 0));
            shouldBeConnected = true;
            firstTransferFinished = false;
            entitiesUpdated = false;
            DataContainer dc = new DataContainer();
            dc.add(name);
            dc.add(level.getChecksum());
            clientImpl = new SimpleClient(server, true, dc);
            clientImpl.addListener(this);
            clientImpl.addPreProcessor(this);
            clientImpl.connect();
            player.activate();
            selectedServer = server;
            respawn();
        } catch (Exception e) {
            if (clientImpl != null) {
                clientImpl.disconnect();
            }
            shouldBeConnected = false;
            clientImpl = null;
            throw e;
        }
    }

    public void disconnect() throws Exception {
        synchronized (SYNC) {
            shouldBeConnected = false;
            coMan.removeAll(world);
            hiView = null;
            if (clientImpl != null) {
                clientImpl.disconnect();
                clientImpl = null;
            }
            player.setAlive();
            player.deactivate();
        }
    }

    public boolean isConnected() {
        return clientImpl != null && clientImpl.isConnected();
    }

    public ServerEntry getServer() {
        return selectedServer;
    }

    public void startServer(int port) throws Exception {
        if (serverImpl == null) {
            serverImpl = new BlueThunderServer(port);
            new Thread(serverImpl).start();
        }
    }

    public void shutDownServer() throws Exception {
        if (serverImpl != null) {
            InfoLine il = new InfoLine(InfoLine.SYSTEM_OUT, 0, "msg", "Server stopped!");
            InfoDataContainer idc = new InfoDataContainer();
            idc.add(il);
            serverImpl.broadcast(idc);
            Thread.sleep(300); // Wait to make sure that the broadcast succeeds!
            serverImpl.stop();
            serverImpl = null;
        }
    }

    public boolean runsServer() {
        return serverImpl != null;
    }

    public void quit() {
        try {
            disconnect();
            shutDownServer();
        } catch (Exception e) {
            // We are on our way to exit...no time for exceptions...
        }
        quit = true;
    }

    public void beforeSending() {
        synchronized (SYNC) {
            prepareClientData();
        }
    }


    public void beforeReceiving() {
        tmpScores = null;
        received = false;
        // Erstmal alle als nicht bertragen markieren. Sollten sie es dennoch werden, wird
        // unten umgesetzt. Ansonsten fliegen sie beim nchsten Zeichnen!
        for (Iterator itty = coMan.getClientObjects().iterator(); itty.hasNext(); ) {
            ClientObject enemy = (ClientObject) itty.next();
            enemy.setRemoteFlag(false);
        }
    }

    public void afterReceiving() {
        if (tmpScores != null) {
            // Highscores sind mitgekommen beim Transfer!
            hiView = new Highscores(tmpScores);
        }
        firstTransferFinished = true;
    }


    public void dataReceivedEnd() {
        received = true;
    }

    public DataContainer[] dataReceived(DataContainer c, int type) {
        if (c != null && c.hasData()) {

            synchronized (SYNC) {
                if (clientImpl != null) {
                    if (type == MessageTypes.OBJ_TRANSFER) {
                        ExtendedDataContainer ec = new ExtendedDataContainer(c);
                        int cnt = 0;
                        while (ec.hasData()) {
                            cnt++;
                            LocalObject lo = ec.getLocalObject();
                            if (lo.getClientID() != clientImpl.getClientID()) {
                                // Somebody else...
                                ClientObject obj = coMan.getOrCreateClientObject(lo, world, bulMan);
                                obj.touch();
                                obj.setToLocalObject(lo);
                                obj.setRemoteFlag(true); // Mark as "belongs to the server"
                            }
                        }
                    }

                    if (type == MessageTypes.EVENT) {
                        EventDataContainer edc = new EventDataContainer(c);
                        int cnt = 0;
                        while (edc.hasData()) {
                            Event event = edc.getEvent();
                            processEvent(event);
                            cnt++;
                        }
                    }

                    if (type == MessageTypes.IMPORTANT_INFO) {
                        InfoDataContainer idc = new InfoDataContainer(c);
                        int cnt = 0;
                        while (idc.hasData()) {
                            InfoLine il = idc.getInfoLine();
                            processInfo(il);
                            cnt++;
                        }
                    }
                }
            }
        }
        return null;
    }

    /**
     * Runs the client.
     * @throws Exception
     */
    private void run() throws Exception {
        Config.saveMemory = true;
        load();
        initWorld();
        initPlayer();
        initClient();
        initGUI();
        gameLoop();
        exit();
    }

    /**
     * Checks for a time out of other clients. If a remote client can't transmit any data, it can't transmit
     * that it's gone either. This is handled here by removing objects from clients that seem to be lost.
     * @throws Exception
     */
    private void checkTimeout() throws Exception {
        if (timeout.getTicks() > 0) {
            Collection objs = coMan.getClientObjects();
            List toKill = new ArrayList();
            for (Iterator itty = objs.iterator(); itty.hasNext(); ) {
                ClientObject co = (ClientObject) itty.next();
                if (co.isOld()) {
                    toKill.add(co);
                }
            }
            for (Iterator itty = toKill.iterator(); itty.hasNext(); ) {
                coMan.remove((ClientObject) itty.next(), world);
            }
        }

        if (clientImpl != null && !clientImpl.isConnected()) {
            // Damn! Where's our connection gone?
            disconnect();
        }
    }

    /**
     * Processes an info message from the server.
     * @param il InfoLine the info
     */
    private void processInfo(InfoLine il) {
        if (il.getType() == InfoLine.PLAYER_ADDED) {
            System.out.println("Player '" + il.getValue() + "' added!");
        }
        if (il.getType() == InfoLine.PLAYER_REMOVED) {
            System.out.println("Player '" + il.getValue() + "' removed!");
        }
        if (il.getType() == InfoLine.SCORE_TABLE) {
            if (tmpScores == null) {
                tmpScores = new HighscoreTable();
            }
            tmpScores.addLineEntry(il);
        }
        if (il.getType() == InfoLine.SYSTEM_OUT) {
            messages.insertCell(2, 0, il.getValue());
        }
    }

    /**
     * Processes an event from the server. An event if something like "player fires a shot" or "player has died".
     * @param event Event the event
     */
    private void processEvent(Event event) {
        switch (event.getType()) {
        case Event.PLAYER_DAMAGE:
            if (event.getTargetClientID() == clientImpl.getClientID()) {
                // Myself
                player.addToValue(event.getValue());
            }
            break;
        case Event.PLAYER_DEAD:
            if (event.getTargetClientID() != clientImpl.getClientID()) {
                // Somebody else
                ClientObject obj = coMan.getClientObject(event.getTargetID(), event.getTargetClientID());
                if (obj != null) {
                    // Can't be null anyway...
                    obj.setCollisionMode(Object3D.COLLISION_CHECK_NONE);
                }
            } else {
                // Myself
                player.setDead();
            }
            break;
        case Event.FIRE:
            if (event.getSourceClientID() != clientImpl.getClientID()) {
                // Someone else has fired!
                System.out.println("BOOM at " + event.getOrigin());
            }
            break;
        case Event.PLAYER_RESET_HEALTH:
            if (event.getSourceClientID() == clientImpl.getClientID()) {
                // Myself...
                player.setValue(100);
            } else {
                ClientObject obj = coMan.getClientObject(event.getSourceID(), event.getSourceClientID());
                if (obj != null) {
                    obj.setCollisionMode(Object3D.COLLISION_CHECK_OTHERS | Object3D.COLLISION_CHECK_SELF);
                }
            }
            break;
        }
    }

    /**
     * Initializes the startup-GUI, i.e. the server browser.
     * @throws Exception
     */
    private void initGUI() throws Exception {
        messages = new Table("msgs", 3, 1, 10, HEIGHT - 60, 500, HEIGHT - 10);
        serverSel = new ServerSelection(browser, this);
        clickMessage = new Label(326, 300);
        clickMessage.setText("Click to respawn!");
        clickMessage.setVisible(false);
    }

    /**
     * Initializes the client.
     */
    private void initClient() {
        coMan = new ClientObjectManager();
        bulMan = new LocalBulletManager();
        timeout = new Ticker(2000);
        fireTicker = new Ticker(200);
        keyStates = new KeyStates();

        browser = new ServerBrowser(7002);
        browser.startBrowser();
    }

    /**
     * Prepares the data from the client that should be send to the server.
     */
    private void prepareClientData() {
        if (clientImpl != null) {
            ExtendedDataContainer playerAndBullets = new ExtendedDataContainer();
            playerAndBullets.add(player);
            synchronized (KILL_SYNC) {
                toKill.addAll(bulMan.fillOrKill(playerAndBullets));
            }
            DataContainer edc = ClientEventQueue.getInstance().getEvents(clientImpl.getClientID());
            ClientEventQueue.getInstance().clear();
            clientImpl.setContainers(new DataContainer[] {playerAndBullets, edc});
        }
    }


    /**
     * Initializes the game world.
     */
    private void initWorld() {
        Config.maxPolysVisible = 65536;
        Config.glVertexArrays = true;
        Config.collideEllipsoidThreshold = 0.01f;
        Config.collideEllipsoidSmoothing = false;
        Config.collideOffset = 150;
        Config.glTrilinear = true;
        Config.glRefresh = 0;
        Config.glZBufferDepth = 16;
        Config.glColorDepth = 24;
        Config.glWindowName = "jPCT, a 3D engine for Java - Visit http://www.jpct.net for more info";

        Logger.setOnError(Logger.ON_ERROR_THROW_EXCEPTION);

        buffer = new FrameBuffer(WIDTH, HEIGHT, FrameBuffer.SAMPLINGMODE_NORMAL);
        buffer.disableRenderer(IRenderer.RENDERER_SOFTWARE);
        buffer.enableRenderer(IRenderer.RENDERER_OPENGL);

        if (!buffer.isInitialized()) {
            // Error with this mode? Try 1024*768 instead...this has to be replaced by a "real" video mode selection
            // in a "real" game.
            buffer = new FrameBuffer(1024, 768, FrameBuffer.SAMPLINGMODE_NORMAL);
            buffer.disableRenderer(IRenderer.RENDERER_SOFTWARE);
            buffer.enableRenderer(IRenderer.RENDERER_OPENGL);
        }

        keyMapper = new KeyMapper();

        world = new World();
        world.addObject(level);
        decals = new DecalManager(world);

        level.enableParticleEmitter(world);

        level.setCollisionOptimization(Object3D.COLLISION_DETECTION_OPTIMIZED);
        level.setCollisionMode(Object3D.COLLISION_CHECK_OTHERS);
        level.addCollisionListener(new LevelEntityListener(decals));
        level.getOcTree().setCollisionUse(OcTree.COLLISION_USE);

        world.setAmbientLight(255, 255, 255);
        world.getLights().setRGBScale(Lights.RGB_SCALE_2X);

        sky = new World();
        sky.addObject(dome);
        sky.setAmbientLight(255, 255, 255);

        SimpleVector center = level.getTransformedCenter();
        center = center.calcSub(dome.getTransformedCenter());
        dome.translate(center);
        dome.setLighting(Object3D.LIGHTING_NO_LIGHTS);
        dome.setAdditionalColor(Color.WHITE);

        mouse = new MouseMapper(buffer);

        crossHair = Primitives.getPlane(1, 2);
        crossHair.setBillboarding(Object3D.BILLBOARDING_ENABLED);
        crossHair.setAdditionalColor(Color.WHITE);
        crossHair.setLighting(Object3D.LIGHTING_NO_LIGHTS);
        crossHair.setTexture("crosshair");
        crossHair.setTransparency(2);
        world.addObject(crossHair);

    }

    /**
     * Initilizes the player. The local player on the client is, opposed to remote players, no real 3D object but
     * just a data representation.
     */
    private void initPlayer() {
        player = new LocalPlayerObject();
        player.setViewRotation(new Matrix());
        player.setRotation(new Matrix());
        player.setSpeed(new SimpleVector());
        player.setType(ClientObjectFactory.PLAYER);
        player.setPosition(spawnPoint);
        player.setDead();
    }

    /**
     * Fires a bullet into the current view direction of the local player.
     */
    private void fire() {
        if ((keyStates.fire || mouse.buttonDown(0)) && clientImpl != null && clientImpl.isConnected()) {
            if (!player.isDead()) {
                player.setAnimation(Animations.FIRE);
                player.setAnimationSpeed(20);
                if (fireTicker.getTicks() > 0) {
                    LocalObject bul = bulMan.addBullet(player, world);
                    Event ev = new Event(Event.FIRE, player, bul);
                    ev.setOrigin(player.getPosition());
                    ClientEventQueue.getInstance().add(ev);
                }
            } else {
                if (clickMessage.isVisible()) {
                    clickMessage.setVisible(false);
                    respawn();
                }
            }
        } else {
            fireTicker.forward();
        }
    }

    /**
     * Respawn!
     */
    private void respawn() {
        // Do this in another thread to make sure, that the server has send its data at least once
        // before respawning.
        entitiesUpdated = false;
        new Thread() {
            public void run() {
                SimpleVector respawn = null;
                int its = 0;
                int cnt = 0;
                try {
                    while (!entitiesUpdated && cnt < 100) {
                        Thread.sleep(100);
                        cnt++;
                    }
                } catch (Exception e) {}

                if (cnt == 100) {
                    throw new RuntimeException("Client: Unable to connect!");
                }

                synchronized (SYNC) {
                    do {
                        SimpleVector s = new SimpleVector(spawnPoints[(int) (Math.random() * spawnPoints.length)]);
                        SimpleVector ss = new SimpleVector();
                        ss.y += 3;
                        if (world.checkCollisionSpherical(s, ss, 5).equals(ss) || its > 15) {
                            respawn = s;
                        } else {
                            its++;
                            System.out.println(s + " is blocked!");
                        }
                    } while (respawn == null);
                    player.setPosition(respawn);
                    player.setAlive();
                }
                clickMessage.setVisible(false);
                Event ev = new Event(Event.PLAYER_RESPAWNED, player, player);
                ClientEventQueue.getInstance().add(ev);
            }
        }.start();
    }

    /**
     * Removes every object from the world that's part of the toKill-List.
     */
    private void cleanUp() {
        synchronized (KILL_SYNC) {
            for (Iterator itty = toKill.iterator(); itty.hasNext(); ) {
                LocalBullet bul = (LocalBullet) itty.next();
                if (bul.getView() != null) {
                    world.removeObject((Object3D) bul.getView());
                }
            }
            toKill.clear();
        }
    }

    /**
     * Updates the particles.
     * @param ticks long the number of ticks passed since last call
     */
    private void updateParticles(long ticks) {
        level.getParticleManager().move(ticks);
    }

    /**
     * Draw the GUI (i.e. the server selection as there is no more GUI in this example) if visible.
     */
    private void drawGUI() {
        if (serverSel.isVisible()) {
            mouse.show();
            serverSel.evaluateInput(mouse, keyMapper);
            serverSel.draw(buffer);
        } else {
            mouse.hide();
        }
        if (keyStates.tab && hiView != null) {
            hiView.draw(buffer);
        }
    }

    /**
     * The game loop. This is where the game...well...loops...:-)
     * @throws Exception
     */
    private void gameLoop() throws Exception {
        Ticker ticker = new Ticker(20);

        blo = new com.threed.jpct.procs.BloomGLProcessor(5, 3, 1, 1);
        buffer.addPostProcessor(blo);
        buffer.clear(new Color(134, 178, 238));

        while (!quit) {
            int ticks = ticker.getTicks();
            if (ticks > 0) {
                if (!serverSel.isVisible()) {
                    keyStates.poll(keyMapper);
                }
                setMode();
                synchronized (SYNC) {
                    updatePlayer(ticks);
                    fire();
                    updateRemoteEntities(ticks);
                    bulMan.moveLocalBullets(ticks, level);
                    updateParticles(ticks);
                    checkTimeout();
                }
                cleanUp();
            }

            buffer.clearZBufferOnly();

            // Note to self: Can't alter farPlanes multiple times during one loop
            // when using a multithreaded renderer...but this is none, so...:-)
            Config.farPlane = 2000;
            sky.renderScene(buffer);
            sky.draw(buffer);

            Config.farPlane = 500;
            world.renderScene(buffer);
            world.draw(buffer);
            if (keyStates.bloom) {
                buffer.runPostProcessors();
            }

            if (clientImpl != null && !clientImpl.isConnected() && shouldBeConnected) {
                messages.insertCell(2, 0, "Disconnected from Server!");
                serverSel.setVisible(true);
                shouldBeConnected = false;
            }

            messages.draw(buffer);
            clickMessage.draw(buffer);
            drawGUI();
            buffer.update();

            buffer.displayGLOnly();
            if (clientImpl != null && clientImpl.isConnected()) {
                Thread.yield();
            } else {
                Thread.sleep(10);
            }
            if (keyStates.exit) {
                serverSel.setVisible(!serverSel.isVisible());
                keyStates.exit = false;
            }
        }
    }

    /**
     * Load resources like textures, the level etc.
     * @throws Exception
     */
    private void load() throws Exception {
        Logger.log("Loading textures...", Logger.MESSAGE);
        SimpleStream ss = new SimpleStream("textures.ser");
        ObjectInput in = new ObjectInputStream(ss.getStream());
        Vector textures = (Vector) in.readObject();
        in.close();
        ss.close();

        TextureManager.getInstance().setState(textures);

        Logger.log("Loading level...", Logger.MESSAGE);
        ss = new SimpleStream("level.ser");
        in = new ObjectInputStream(ss.getStream());
        Object3D tmp = (Object3D) in.readObject();
        level = new Level(tmp);
        level.setOcTree(tmp.getOcTree());
        in.close();
        ss.close();

        ss = new SimpleStream("sky.jpg");
        TextureManager.getInstance().addTexture("sky", new Texture(ss.getStream()));
        ss.close();

        ss = new SimpleStream("particle.jpg");
        TextureManager.getInstance().addTexture("particle", new Texture(ss.getStream()));
        ss.close();

        ss = new SimpleStream("crosshair.gif");
        TextureManager.getInstance().addTexture("crosshair", new Texture(ss.getStream()));
        ss.close();

        ss = new SimpleStream("decal.gif");
        TextureManager.getInstance().addTexture("decal", new Texture(ss.getStream()));
        ss.close();

        ss = new SimpleStream("dome.3ds");
        dome = Loader.load3DS(ss.getStream(), 1f)[0];
        dome.setTexture("sky");
        dome.build();
        dome.translate(0, -330, 0);
        dome.translateMesh();
        dome.rotateX( -(float) Math.PI / 2f);
        dome.calcTextureWrap();
        ss.close();
    }

    /**
     * Terminate everything including the VM.
     * @throws Exception
     */
    private void exit() throws Exception {
        if (clientImpl != null) {
            clientImpl.disconnect();
        }
        buffer.disableRenderer(IRenderer.RENDERER_OPENGL);
        buffer.dispose();
        Thread.sleep(500);
        System.exit(0);
    }

    /**
     * Updates all entities (like player meshes) that are remote, i.e. that came from the server.
     * @param ticks long the number of ticks passed
     */
    private void updateRemoteEntities(long ticks) {
        List toKick = new ArrayList();
        for (Iterator itty = coMan.getClientObjects().iterator(); itty.hasNext(); ) {
            ClientObject enemy = (ClientObject) itty.next();
            if (enemy.isRemote()) {
                enemy.process(ticks, level);
            } else {
                if (received) {
                    toKick.add(enemy);
                }
            }
        }
        for (Iterator itty = toKick.iterator(); itty.hasNext(); ) {
            coMan.remove((ClientObject) itty.next(), world);
        }
        if (firstTransferFinished) {
            entitiesUpdated = true;
        }
    }

    /**
     * Updates the crosshair.
     */
    private void setCrossHair() {
        SimpleVector pos = bulMan.getInitialPosition(player);
        Matrix rot = player.getViewRotation().invert3x3();
        SimpleVector speed = rot.getZAxis();
        float d = world.calcMinDistance(pos, speed, 10000);
        if (d == Object3D.COLLISION_NONE) {
            d = 100;
        }
        speed.scalarMul(d - 2f);
        speed.add(pos);
        crossHair.setOrigin(speed);
    }

    /**
     * Updates the local player. Albeit the local player is no mesh, this method takes care of setting the correct
     * animation because that will be transfered to the server and all other clients.
     * @param ticks long the number of ticks passed
     */
    private void updatePlayer(long ticks) {
        Camera camera = world.getCamera();
        if (clientImpl != null && clientImpl.isConnected()) {
            if (!player.isDead()) {
                boolean moved = player.move(world, keyStates, mouse, ticks);
                setCrossHair();
                if (moved) {
                    player.setAnimation(Animations.MOVE);
                    player.setAnimationSpeed(40);
                } else {
                    player.setAnimation(Animations.NONE);
                    player.setAnimationSpeed(120);
                }
            } else {
                if (System.currentTimeMillis() - player.deadSince() > 1000) {
                    player.setAnimation(Animations.DEAD);
                    if (System.currentTimeMillis() - player.deadSince() > 5000) {
                        clickMessage.setVisible(true);
                    }
                } else {
                    player.setAnimation(Animations.DIE);
                }
                player.setAnimationSpeed(80);
            }
        }
        player.alignCamera(camera);
        sky.getCamera().setBack(camera.getBack().cloneMatrix());
    }

    /**
     * Sets the sampling mode. Can be changed in game by pressing 1-3 if the hardware supports it.
     */
    private void setMode() {
        if (keyStates.one && videoMode != FrameBuffer.SAMPLINGMODE_HARDWARE_ONLY) {
            setMode(FrameBuffer.SAMPLINGMODE_HARDWARE_ONLY);
            keyStates.one = false;
        }
        if (keyStates.two && videoMode != FrameBuffer.SAMPLINGMODE_GL_AA_2X) {
            setMode(FrameBuffer.SAMPLINGMODE_GL_AA_2X);
            keyStates.two = false;
        }
        if (keyStates.three && videoMode != FrameBuffer.SAMPLINGMODE_GL_AA_4X) {
            setMode(FrameBuffer.SAMPLINGMODE_GL_AA_4X);
            keyStates.three = false;
        }
    }

    private void setMode(int mode) {
        videoMode = mode;
        buffer.removePostProcessor(blo);
        keyMapper.destroy();
        buffer.disableRenderer(IRenderer.RENDERER_OPENGL);
        buffer.dispose();
        buffer = new FrameBuffer(WIDTH, HEIGHT, mode);
        buffer.disableRenderer(IRenderer.RENDERER_SOFTWARE);
        buffer.enableRenderer(IRenderer.RENDERER_OPENGL);
        keyMapper = new KeyMapper();
        buffer.addPostProcessor(blo);
    }
}
