package feud.testgame;

import java.util.*;

import feud.clientserver.*;
import feud.testgame.model.*;
import feud.testgame.util.*;

/**
 * This is this test game's server component. It's usually started from within a running client but can run as
 * a stand-alone server as well.
 */
public class BlueThunderServer implements DataTransferListener, Runnable, ClientLoginListener, ClientLogoutListener {

    private final static Object SYNC=new Object();

    private boolean exit = false;

    private SimpleServer serverImpl=null;
    private ExtendedDataContainer serverState=null;
    private ServerObjectManager serverObjMan=null;
    private ServerEventManager eventMan=null;

    private int port=0;
    private Map client2Info=new HashMap();
    private int lastLevelChksum=-9999;

    /**
     * Starts a new server using the default tcp-port.
     */
    public BlueThunderServer() {
        this(SimpleServer.DEFAULT_PORT);
    }

    /**
     * Starts a new server using a specified tcp-port.
     * @param port int the port
     */
    public BlueThunderServer(int port) {
        this.port=port;
    }

    /**
     * Run the server as stand-alone.
     * @param args String[]
     * @throws Exception
     */
    public static void main(String[] args) throws Exception {
        BlueThunderServer str = new BlueThunderServer();
        str.run();
    }

    public DataContainer loggedOut(ClientInfo ci) {
        InfoDataContainer dc=new InfoDataContainer();

        PlayerInfo pi=(PlayerInfo) client2Info.get(new Integer(ci.getID()));
        if (pi!=null) {
            dc=new InfoDataContainer();
            InfoLine il=new InfoLine(InfoLine.PLAYER_REMOVED, 0, "name", pi.getName());
            dc.add(il);
            il=new InfoLine(InfoLine.SYSTEM_OUT, 0, "msg", "Player '"+pi.getName()+"' has left the game!");
            dc.add(il);
        } else {
            throw new RuntimeException("?????? "+client2Info.size());
        }

        serverObjMan.remove(ci);
        client2Info.remove(new Integer(ci.getID()));

        addScores(dc);

        return dc;
    }

    public DataContainer loggedIn(ClientInfo ci, DataContainer dc) {
        String playerName="player";
        if (dc!=null) {
            String tmp=dc.getNextString();
            int chk=dc.getNextInt();
            if (tmp!=null && tmp.length()>0) {
                playerName = tmp;
            }
            // very simple way to prove that the level is the same on all clients.
            // the first one with a different level will be rejected!
            if (lastLevelChksum!=-9999 && chk!=lastLevelChksum) {
                return null;
            }
            lastLevelChksum=chk;
        }
        PlayerInfo pi=new PlayerInfo(playerName);
        client2Info.put(new Integer(ci.getID()), pi);
        String txt="Player '"+playerName+"' has joined the game!";
        NetLogger.log(txt);
        InfoDataContainer ic=new InfoDataContainer();
        InfoLine il=new InfoLine(InfoLine.PLAYER_ADDED, 0, "name", playerName);
        ic.add(il);
        il=new InfoLine(InfoLine.SYSTEM_OUT, 0, "msg", txt);
        ic.add(il);
        addScores(ic);
        return ic;
    }

    public void run() {
        try {
            initServer();
            runServer();
            stop();
        } catch(Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Stops the server.
     */
    public void stop()  {
        exit=true;
        serverImpl.shutDown();
    }

    public void dataReceivedEnd() {}

    /**
     * Gets the player info a client's ID.
     * @param clientID int the ID
     * @return PlayerInfo the player info
     */
    public PlayerInfo getPlayerInfo(int clientID) {
        return (PlayerInfo) client2Info.get(new Integer(clientID));
    }

    /**
     * Broadcasts data to all clients.
     * @param dc DataContainer the container with the data to broadcast
     */
    public void broadcast(DataContainer dc) {
        serverImpl.broadcast(dc);
    }

    /**
     * Removes an object from the server
     * @param loID int
     * @param clientID int
     * @param ci ClientInfo
     */
    public void remove(int loID, int clientID, ClientInfo ci) {
        synchronized(SYNC) {
            LocalObject lo=serverObjMan.getLocalObjectToIDs(loID, clientID);
            serverObjMan.removeObject(lo, ci);
        }
    }

    public DataContainer[] dataReceived(DataContainer c, int type) {
       try {
           if (c != null) {
               synchronized(SYNC) {
                   ExtendedDataContainer ec = new ExtendedDataContainer(c);
                   if (type==MessageTypes.OBJ_TRANSFER) {
                       while (ec.hasData()) {
                           serverObjMan.setOrUpdate(ec);
                       }
                       return new DataContainer[] {(DataContainer) serverState.clone()};
                   }
                   if (type==MessageTypes.EVENT) {
                       EventDataContainer edc=new EventDataContainer(ec);
                       List resp=new ArrayList();
                       ClientInfo ci=edc.getClientInfo();
                       while (edc.hasData()) {
                           DataContainer[] res=eventMan.manageEvent(edc.getEvent(), serverObjMan, this, ci);
                           if (res!=null) {
                               resp.addAll(Arrays.asList(res));
                           }
                       }
                       if (resp.size()==0) {
                           return null;
                       }
                       DataContainer[] res=new DataContainer[resp.size()];
                       for (int i=0; i<res.length; i++) {
                           res[i]=(DataContainer) resp.get(i);
                       }
                       return res;
                   }
               }
           }
       } catch(Exception e) {
           throw new RuntimeException(e);
       }
       return null;
    }

    /**
     * Adds the scores to a highscore table (a GUI-element).
     * @param idc InfoDataContainer the info container containing the scores
     */
    public void addScores(InfoDataContainer idc) {
        HighscoreTable hi = new HighscoreTable();
        for (Iterator itty = client2Info.values().iterator(); itty.hasNext(); ) {
            PlayerInfo pi = (PlayerInfo) itty.next();
            hi.addLine(pi);
        }
        hi.addToContainer(idc);
    }

    /**
     * Initialize the server
     */
    private void initServer() {
        serverState=new ExtendedDataContainer();
        String name=System.getProperty("user.name");
        if (name==null) {
            name="My server";
        } else {
            if (name.endsWith("s")) {
                name+="'";
            } else {
                name+="'s";
            }
            name+=" server";
        }
        serverImpl=new SimpleServer(port,SimpleServer.UDP_DEFAULT_PORT,true, name);
        serverImpl.addListener(this);
        serverObjMan=new ServerObjectManager();
        serverImpl.addLogoutListener(this);
        serverImpl.addLoginListener(this);
        eventMan=new ServerEventManager();
    }

    /**
     * Creates a new "server state". This means that all the data living on the server gets collected
     * and stored in a structure ready to be transfered to the clients.
     */
    private void createServerState() {
        synchronized(SYNC) {
            serverState=new ExtendedDataContainer();
            // Transfer all client data
            serverObjMan.fill(serverState);
        }
    }

    /**
     * The server's main loop.
     * @throws Exception
     */
    private void runServer() throws Exception {
        Ticker ticker = new Ticker(20);

        while (!exit) {
            int ticks = ticker.getTicks();
            if (ticks > 0) {
                createServerState();
            } else {
                Thread.sleep(10);
            }
        }
    }
}
