package feud.testgame.view;

import com.threed.jpct.*;

/**
 * A simple manager for particles. I wouldn't call it a "system" for now...
 */
public class ParticleManager {

  private Particle[] parts = null;
  private World level=null;
  private long lastCleanup=0;
  private int cnt=100;

  /**
   * Creates a new particle manager with a default number of particles (100).
   * @param level World the world
   */
  public ParticleManager(World level) {
    this.level=level;
    init();
  }

  /**
   * Creates a new particle manager for a number of particles.
   * @param level World the world
   * @param max int the number of particles
   */
  public ParticleManager(World level, int max) {
     cnt=max;
     this.level=level;
     init();
  }

  /**
   * Moves all particles manmaged by this manager.
   * @param ticks long the number of ticks passed since the last call
   */
  public void move(long ticks) {
    for (int i = 0; i < cnt; i++) {
      Particle pp = parts[i];
      if (pp!=null && pp.getVisibility()) {
        pp.move(ticks);
      }
    }
    doCleanup();
  }

  /**
   * Adds a new particle to the manager. If the maximum number of active, visible particles have been reached, nothing happens.
   * This method is for adding a particle that it limited in y-direction but not in time.
   * @param pos SimpleVector the position
   * @param vel SimpleVector the velocity
   * @param limit float the y-limit
   * @param texture String the texture's name
   */
  public void addParticle(SimpleVector pos, SimpleVector vel, float limit, String texture) {
    addParticle(pos, vel, limit, -1, texture);
  }

  /**
   * Adds a new particle to the manager. If the maximum number of active, visible particles have been reached, nothing happens.
   * This method is for adding a particle that it in time but not in y-direction.
   * @param pos SimpleVector the position
   * @param vel SimpleVector the velocity
   * @param time long the life time
   * @param texture String the texture's name
   */
  public void addParticle(SimpleVector pos, SimpleVector vel, long time, String texture) {
     addParticle(pos, vel, 9999999999999999f, time, texture);
  }

  /**
   * Adds a new particle to the manager. If the maximum number of active, visible particles have been reached, nothing happens.
   * The particle will be limited in time and y-direction.
   * @param pos SimpleVector the position
   * @param vel SimpleVector the velocity
   * @param limit float the y-limit
   * @param time long the time-limit
   * @param texture String the texture's name
   */
  public void addParticle(SimpleVector pos, SimpleVector vel, float limit, long time, String texture) {
    Particle p = getParticle();
    if (p!=null) {
       p.setOrigin(pos);
       p.setVelocity(vel);
       p.setLifeTime(time);
       p.setYLimit(limit);
       p.setTexture(texture);
       p.reset();
    }
  }

  /**
   * Initializes the manager
   */
  private void init() {
     parts=new Particle[cnt];
     lastCleanup=System.currentTimeMillis();
  }

  /**
   * Every 30 secs, the managers checks for old particles that can be cleaned up.
   */
  private void doCleanup() {
     if (System.currentTimeMillis()-lastCleanup>30000) {
        int cnt=0;
        for (int i=cnt-1; i>=0; i--) {
           Particle pp=parts[i];
           if (pp!=null && pp.isOld()) {
              level.removeObject(pp);
              parts[i]=null;
              cnt++;
           }
        }
        if (cnt!=0) {
           Logger.log("Particle system cleanup performed: "+cnt+" particles removed", Logger.MESSAGE);
        }
        lastCleanup=System.currentTimeMillis();
     }
  }

  /**
   * Gets a particle from the manager's pool. If the pool if full of particles, null will be returned.
   * @return Particle a fresh particle or null
   */
  private Particle getParticle() {
    for (int i = 0; i < cnt; i++) {
      Particle pp = parts[i];
      if (pp!=null && !pp.getVisibility()) {
        pp.setVisibility(Object3D.OBJ_VISIBLE);
        return pp;
      }
    }
    Particle p=new Particle();
    level.addObject(p);
    boolean found=false;
    for (int i=0; i<cnt; i++) {
       if (parts[i]==null) {
          parts[i]=p;
          found=true;
          break;
       }
    }
    if (!found) {
       return null;
    }
    return p;
  }
}
