package feud.testgame;

import feud.testgame.model.*;
import com.threed.jpct.*;

/**
 * An event is everything that the server and the other client should be notified of but that's not really physical, like
 * a new position of the player would be.
 */
public class Event {

    public static final int BULLET_HIT = 1;
    public static final int PLAYER_DAMAGE = 2;
    public static final int PLAYER_DEAD = 3;
    public final static int ENTITY_REMOVE=4;
    public final static int FIRE=8;
    public final static int PLAYER_RESPAWNED=16;
    public final static int PLAYER_RESET_HEALTH=32;

    private int sourceID = -1;
    private int targetID = -1;
    private int targetClientID = -1;
    private int sourceClientID = -1;
    private int type = 0;
    private int value = 0;
    private SimpleVector origin=SimpleVector.ORIGIN;

    /**
     * Creates a new event. This is used for creating bullet hit events.
     * @param type int the type of the event. See the constants above of possible types.
     * @param sourceID int the object id of the source object triggering this event (if any)
     * @param targetID int the id of the target object of this event (if any).
     * @param targetClientID int the id of the target client
     */
    public Event(int type, int sourceID, int targetID, int targetClientID) {
        this.sourceID = sourceID;
        this.targetID = targetID;
        this.targetClientID = targetClientID;
        this.type = type;
    }

    /**
     * Creates a new event.
     * @param type int the type of the event. See the constants above of possible types.
     * @param source LocalObject the source object of the event
     * @param target LocalObject the target object of the event
     */
    public Event(int type, LocalObject source, LocalObject target) {
        this(type, source.getObjectID(), target.getObjectID(), target.getClientID());
        setSourceClientID(source.getClientID());
    }

    /**
     * Sets the origin of the event. Not all events have this.
     * @param org SimpleVector the origin
     */
    public void setOrigin(SimpleVector org) {
        origin=org;
    }

    /**
     * Gets the origin of th event. Not all events have this.
     * @return SimpleVector the origin
     */
    public SimpleVector getOrigin() {
        return origin;
    }

    /**
     * Sets a value for the event. Not all events have this.
     * @param val int the value
     */
    public void setValue(int val) {
        this.value = val;
    }

    /**
     * Returns the value for the event. Not all events have this.
     * @return int the value
     */
    public int getValue() {
        return value;
    }

    /**
     * Returns the type of the event.
     * @return int the type
     */
    public int getType() {
        return type;
    }

    /**
     * Sets the id of the source client.
     * @param id int the id
     */
    public void setSourceClientID(int id) {
        this.sourceClientID = id;
    }

    /**
     * Returns the id of the source client.
     * @return int the id
     */
    public int getSourceClientID() {
        return sourceClientID;
    }

    /**
     * Returns the id of the source object.
     * @return int the id
     */
    public int getSourceID() {
        return sourceID;
    }

    /**
     * Returns the id of the target object.
     * @return int the id
     */
    public int getTargetID() {
        return targetID;
    }

    /**
     * Returns the id of the target client
     * @return int the if
     */
    public int getTargetClientID() {
        return targetClientID;
    }

    public String toString() {
        return this.getType() + "/" + this.getSourceID() + "/" + this.getSourceClientID() + "/" + this.getTargetID() + "/" + this.getTargetClientID();
    }
}
