package feud.testgame;

import feud.testgame.model.*;
import feud.clientserver.*;

/**
 * A helper class for the server to ease event processing. An event is send from the client to the server and processed there.
 * This is what this class does.
 */
public class ServerEventManager {

    /**
     * Manages an event.
     * @param event Event the event to manage
     * @param servMan ServerObjectManager the server object manager that lives on this server
     * @param server BlueThunderServer the actual server instance
     * @param ci ClientInfo the client info of the client from which the event came from
     * @return DataContainer[] an optional response. May be null (and usually is in the current implementation).
     */
    public DataContainer[] manageEvent(Event event, ServerObjectManager servMan, BlueThunderServer server, ClientInfo ci) {
        switch (event.getType()) {
        case Event.BULLET_HIT:
            return processBulletHit(event, servMan, server);
        case Event.ENTITY_REMOVE:
            return processEntityRemoval(event, ci, server);
        case Event.FIRE:
            return processFire(event, server);
        case Event.PLAYER_RESPAWNED:
            return processRespawn(event, servMan, server);
        default:
            return null;
        }
    }

    /**
     * Process a players respawn.
     * @param event Event the event
     * @param servMan ServerObjectManager the server object manager that lives on this server
     * @param server BlueThunderServer the actual server instance
     * @return DataContainer[] an optional response. May be null.
     */
    private DataContainer[] processRespawn(Event event, ServerObjectManager servMan, BlueThunderServer server) {
        LocalObject source = servMan.getLocalObjectToIDs(event.getSourceID(), event.getSourceClientID());
        source.setValue(100); // Health to max!
        EventDataContainer edc = new EventDataContainer();
        Event ev = new Event(Event.PLAYER_RESET_HEALTH, source, source);
        edc.add(ev);
        server.broadcast(edc);
        PlayerInfo sip = server.getPlayerInfo(event.getSourceClientID());
        InfoDataContainer idc = new InfoDataContainer();
        InfoLine il = new InfoLine(InfoLine.SYSTEM_OUT, 0, "msg", "Player '" + sip.getName() + "' has spawned!");
        idc.add(il);
        server.broadcast(idc);
        return null;
    }

    /**
     * Process firing.
     * @param event Event the event
     * @param server BlueThunderServer the actual server instance
     * @return DataContainer[] an optional response. May be null.
     */
    private DataContainer[] processFire(Event event, BlueThunderServer server) {
        EventDataContainer edc = new EventDataContainer();
        edc.add(event);
        server.broadcast(edc);
        return null;
    }

    /**
     * Process a client's request to remove an entity.
     * @param event Event the event
     * @param ci ClientInfo the client info
     * @param server BlueThunderServer the server instance
     * @return DataContainer[] an optional response. May be null.
     */
    private DataContainer[] processEntityRemoval(Event event, ClientInfo ci, BlueThunderServer server) {
        server.remove(event.getSourceID(), event.getSourceClientID(), ci);
        return null;
    }

    /**
     * Process a bullet hit.
     * @param event Event the vent
     * @param servMan ServerObjectManager the server object manager
     * @param server BlueThunderServer the server instance
     * @return DataContainer[] an optional response. May be null.
     */
    private DataContainer[] processBulletHit(Event event, ServerObjectManager servMan, BlueThunderServer server) {
        LocalObject source = servMan.getLocalObjectToIDs(event.getSourceID(), event.getSourceClientID()); // Die Quelle ist immer vom Client...
        LocalObject target = servMan.getLocalObjectToIDs(event.getTargetID(), event.getTargetClientID());

        if (source == null || target == null) {
            throw new RuntimeException("Server can't find client object for: " + event + "/" + source + "/" + target);
        }

        // Sanity check to see if this is actually possible. Because the clients do these checks, there's room for cheating.
        // You may prevent this (i needed) by doing a better check or by doing all checks on the server alone...this depends
        // on your needs and the application. In this "game", we almost trust the client...albeit you never should...
        if (source.getPosition().calcSub(target.getPosition()).length() < 10) {
            if (target.getValue() > 0) { // Lebt der noch?
                target.addToValue( -20);
                EventDataContainer edc = new EventDataContainer();
                Event eve = new Event(Event.PLAYER_DAMAGE, source, target); //Schaden!
                eve.setSourceClientID(source.getClientID());
                eve.setValue( -20);
                edc.add(eve);
                if (target.getValue() <= 0) {
                    Event eve2 = new Event(Event.PLAYER_DEAD, source, target); // Tot!
                    eve.setSourceClientID(source.getClientID());
                    edc.add(eve2);
                    PlayerInfo sip = server.getPlayerInfo(event.getSourceClientID());
                    PlayerInfo tip = server.getPlayerInfo(event.getTargetClientID());
                    sip.incFrags();
                    tip.incKilled();
                    InfoDataContainer idc = new InfoDataContainer();
                    server.addScores(idc);
                    InfoLine il = new InfoLine(InfoLine.SYSTEM_OUT, 0, "msg", "'" + sip.getName() + "' has killed '" + tip.getName() + "'!");
                    idc.add(il);
                    server.broadcast(idc); // Highscores!
                }
                server.broadcast(edc);
            }
        }
        return null;
    }
}
