package feud.testgame.model;

import java.util.*;

import com.threed.jpct.*;
import feud.testgame.*;
import feud.testgame.view.*;

/**
 * Manager class for bullets. A bullet is an instance of LocalBullet in this case.
 */
public class LocalBulletManager {

    private List bullets=new ArrayList();

    /**
     * Creates a new manager.
     */
    public LocalBulletManager() {
    }

    /**
     * The collision detection routines work with the view objects, not with the model. Therefore,
     * it's required to map the view to the model. To minimize dependencies, this is done by using this
     * mathod.
     * @param co the view
     * @return LocalObject the corresponding model (or null if there is none...what shouldn't happen)
     */
    public LocalObject getLocalObjectToView(ClientObject co) {
        for (Iterator itty=bullets.iterator(); itty.hasNext();) {
            LocalBullet lob=(LocalBullet) itty.next();
            if (co.equals(lob.getView())) {
                return lob;
            }
        }
        return null;
    }

    /**
     * Gets the initial position of a newly fired bullet.
     * @param player the player which fires the bullet
     * @return SimpleVector the position
     */
    public SimpleVector getInitialPosition(LocalPlayerObject player) {
        SimpleVector pos = new SimpleVector(player.getPosition());
        Matrix rot = player.getViewRotation().invert3x3();
        SimpleVector speed = rot.getZAxis();
        speed.scalarMul(2);
        SimpleVector addy = rot.getYAxis();
        addy.scalarMul(3.5f);
        pos.add(addy);
        addy = rot.getXAxis();
        pos.add(addy);
        return pos;
    }

    /**
     * Adds a new bullet to the manager and, if not null, a new view for that bullet to the world.
     * @param player the player which fires the bullet
     * @param world the world in which the player exists
     * @return LocalObject the new bullet model
     */
    public LocalObject addBullet(LocalPlayerObject player, World world) {
        SimpleVector pos=getInitialPosition(player);
        Matrix rot=player.getViewRotation().invert3x3();
        SimpleVector speed=rot.getZAxis();
        speed.scalarMul(1.5f);

        LocalBullet bul=new LocalBullet();
        bul.setPosition(new SimpleVector(pos));
        bul.setRotation(rot.cloneMatrix());
        bul.setSpeed(speed);
        bul.setType(ClientObjectFactory.BULLET);

        if (world!=null) {
            ClientObject b=ClientObjectFactory.getInstance().create(ClientObjectFactory.BULLET);
            b.setToLocalObject(bul);
            b.addToWorld(world);
            bul.setView(b);
            b.getTranslationMatrix().setIdentity();
            b.translate(bul.getPosition());
            b.addCollisionListener(new BulletCollisionListener(bul));
        }
        bullets.add(bul);
        return bul;
    }

    /**
     * Moves the bullets according top their speed. This happens to the model and, if present, to the view too.
     * @param ticks the game time that has passed since the last call
     * @param level the current level (this can be null if the bullet has no view)
     */
    public void moveLocalBullets(long ticks, Level level) {
        ParticleManager parMan=level.getParticleManager();
        for (Iterator itty=bullets.iterator(); itty.hasNext();) {
            LocalBullet bul=(LocalBullet) itty.next();
            bul.move(ticks);

            ClientObject bv=bul.getView();
            if (bv!=null) {
                bv.processLocal(ticks, level, bul);
            }
        }
        parMan.move(ticks);
    }

    /**
     * Prepares a data container with the data of all currently active bullets on a client and adds
     * all still exisiting but inactive bullets to a "to-kill"-List.
     * @param dc the container to be filled
     * @return List the list with inactive bullets that weren't added to the container
     */
    public List fillOrKill(ExtendedDataContainer dc) {
        List toKill=new ArrayList();
        for (Iterator itty=bullets.iterator(); itty.hasNext();) {
            LocalBullet bul=(LocalBullet) itty.next();
            if (bul.isOld() || bul.hasCollided()) {
                itty.remove();
                toKill.add(bul);
                Event ev=new Event(Event.ENTITY_REMOVE, bul, bul);
                ClientEventQueue.getInstance().add(ev);
            }
            dc.add(bul);
        }
        return toKill;
    }
}
