package feud.testgame.model;

import com.threed.jpct.*;
import feud.testgame.*;
import feud.testgame.util.*;

/**
 * A local object that holds a player's data.
 */
public class LocalPlayerObject extends LocalObject {

    private Matrix viewRot = new Matrix();
    private SimpleVector ellipsoid = new SimpleVector(3, 5, 3);
    private float turnSpeed = (float) Math.toRadians( -2f);

    private float moveSpeed = 0.75f;
    private boolean isDead = false;
    private long time = 0;
    private boolean playingDead = false;
    private boolean isActive = false;
    private float xAngle = 0;

    /**
     * Gets the "view rotation". This is the direction in which the player is looking. Not the direction
     * in which he/she's moving.
     * @return Matrix the view rotation
     */
    public Matrix getViewRotation() {
        return viewRot;
    }

    /**
     * Sets the "view rotation". This is the direction in which the player is looking. Not the direction
     * in which he/she's moving.
     * @param mat the view direction
     */
    public void setViewRotation(Matrix mat) {
        viewRot = mat;
    }

    /**
     * Flag the player as dead.
     */
    public void setDead() {
        if (!isDead) {
            playingDead = false;
            isDead = true;
            time = System.currentTimeMillis();
        }
    }

    /**
     * Flag the player as alive.
     */
    public void setAlive() {
        if (isDead) {
            isDead = false;
            time = 0;
            playingDead = false;
        }
    }

    /**
     * Is the player dead?
     * @return boolean is he/she?
     */
    public boolean isDead() {
        return isDead;
    }

    /**
     * Is the player is dead, here's since how long...
     * @return long how long is the player dead?
     */
    public long deadSince() {
        return time;
    }

    /**
     * Activate a player. An active player is currently playing but can very well be dead.
     */
    public void activate() {
        isActive = true;
    }

    /**
     * Deactivate a player.
     */
    public void deactivate() {
        isActive = false;
    }


    /**
     * Is the player active. An active player is currently playing but can very well be dead.
     * @return boolean is the player active?
     */
    public boolean isActive() {
        return isActive;
    }

    /**
     * Align the camera to the player.
     * @param camera the camera that should be aligned
     */
    public void alignCamera(Camera camera) {
        if (getPosition() != null) {
            if (!playingDead && isActive()) {
                camera.rotateX((float) Math.random() - 0.5f);
                camera.rotateY((float) Math.random() - 0.5f);
                camera.rotateZ((float) Math.random() - 0.5f);
                playingDead = true;
            } else {
                if (!isDead() || !isActive()) {
                    SimpleVector pos = getPosition();
                    camera.setPosition(pos);
                    camera.setBack(getViewRotation().cloneMatrix());
                }
            }
        }
    }

    /**
     * Move the player according to key and mouse input.
     * @param world the world in which the player exists
     * @param keys the currently pressed keys
     * @param mouse the mouse input
     * @param ticks the game time passed since the last call
     * @return boolean true, if the player has moved. False otherwise.
     */
    public boolean move(World world, KeyStates keys, MouseMapper mouse, long ticks) {
        SimpleVector grav = new SimpleVector(0, ticks, 0);
        SimpleVector pos = getPosition();
        boolean changed = false;
        if (pos != null) {
            pos.add(new SimpleVector(0, ellipsoid.y, 0));
            SimpleVector dir = world.checkCollisionEllipsoid(pos, grav, ellipsoid, 1);
            pos.add(new SimpleVector(0, -ellipsoid.y, 0));
            dir.x = 0;
            dir.z = 0;
            pos.add(dir);
            setSpeed(dir);

            pos.add(new SimpleVector(0, ellipsoid.y, 0));

            if (keys.up) {
                changed = true;
                SimpleVector temp = getRotation().getZAxis();
                temp.scalarMul(moveSpeed * (float) ticks);
                temp = world.checkCollisionEllipsoid(pos, temp, ellipsoid, 8);
                pos.add(temp);
                temp.add(dir);
                setSpeed(new SimpleVector(temp));
            }

            if (keys.down) {
                changed = true;
                SimpleVector temp = getRotation().getZAxis();
                temp.scalarMul( -moveSpeed * (float) ticks);
                temp = world.checkCollisionEllipsoid(pos, temp, ellipsoid, 8);
                pos.add(temp);
                temp.add(dir);
                setSpeed(new SimpleVector(temp));
            }

            if (keys.left) {
                changed = true;
                SimpleVector temp = getRotation().getXAxis();
                temp.scalarMul( -moveSpeed * (float) ticks);
                temp = world.checkCollisionEllipsoid(pos, temp, ellipsoid, 8);
                pos.add(temp);
                temp.add(dir);
                setSpeed(new SimpleVector(temp));
            }

            if (keys.right) {
                changed = true;
                SimpleVector temp = getRotation().getXAxis();
                temp.scalarMul(moveSpeed * (float) ticks);
                temp = world.checkCollisionEllipsoid(pos, temp, ellipsoid, 8);
                pos.add(temp);
                temp.add(dir);
                setSpeed(new SimpleVector(temp));
            }

            Matrix rot = getRotation();

            int dx = mouse.getDeltaX();
            int dy = mouse.getDeltaY();

            float ts = turnSpeed * (float) ticks;
            float tsy = ts;

            if (dx != 0) {
                ts = Math.abs(dx) / -500f;
            }
            if (dy != 0) {
                tsy = Math.abs(dy) / 500f;
            }

            if (dx < 0) {
                viewRot.rotateAxis(viewRot.getYAxis(), ts);
                rot.rotateY(ts);
            }

            if (dx > 0) {
                viewRot.rotateAxis(viewRot.getYAxis(), -ts);
                rot.rotateY( -ts);
            }

            if (keys.lookUp || dy > 0) {
                if (xAngle < Math.PI / 4.2) {
                    viewRot.rotateX(tsy);
                    xAngle += tsy;
                }
            }
            if (keys.lookDown || dy < 0) {
                if (xAngle > -Math.PI / 4.2) {
                    viewRot.rotateX( -tsy);
                    xAngle -= tsy;
                }
            }

            pos.add(new SimpleVector(0, -ellipsoid.y, 0));
            setPosition(pos);
        }
        return changed;
    }
}
