package feud.testgame.view;

import com.threed.jpct.*;
import feud.testgame.model.*;

/**
 * A view object that represents a bullet.
 */
public class BulletView extends ClientObject {

    private static Object3D bluePrint = null;

    static {
        // Static initializer for bullet blueprint
        bluePrint = Primitives.getSphere(0.2f);
        bluePrint.getMesh().compress();
        bluePrint.build();
    }

    /**
     * Creates a new bullet based on the blueprint.
     */
    public BulletView() {
        super(bluePrint);
        this.setCollisionMode(Object3D.COLLISION_CHECK_SELF);
    }

    /**
     * Creates a new bullet based on another Object3D.
     * @param obj the base object
     */
    public BulletView(Object3D obj) {
        super(obj);
        this.setCollisionMode(Object3D.COLLISION_CHECK_SELF);
    }

    /**
     * Processes local bullets, i.e. bullets that have been fired on this client. Only those bullets may
     * be checked for collisions. All the others will be checked by theirs creating clients and verified by
     * the server.
     * @param ticks time passed since the last run
     * @param level the current level
     * @param lo the LocalObject that this client object represents. Client objects and local objects are not
     * bound directly, which is why the client object doesn't know of its local one.
     */
    public void processLocal(long ticks, Level level, LocalObject lo) {
        if (!lo.hasCollided()) {
            setToLocalObject(lo);
            SimpleVector pos = new SimpleVector(getBackPosition());

            getRotationMatrix().setTo(lo.getSpeed().getRotationMatrix());

            ParticleManager parMan = level.getParticleManager();

            if (parMan != null) {
                if (Math.random() < 0.5f * ticks) {
                    SimpleVector vel = lo.getSpeed().normalize();
                    vel.x += 0.05 - Math.random() / 10f;
                    vel.y += 0.05 - Math.random() / 10f;
                    vel.z += 0.05 - Math.random() / 10f;
                    vel.scalarMul( -1.1f);
                    parMan.addParticle(pos, vel, 1000L, "particle");
                }
            }

            SimpleVector tc = getTransformedCenter();
            pos = pos.calcSub(tc);
            float len = pos.length();
            int id = this.checkForCollision(pos.normalize(), len);
            if (id == Object3D.NO_OBJECT) {
                translate(pos);
            }
            setModified(true);
        }
    }

    /**
     * Process remote bullets.
     * @param ticks time passed since the last run
     * @param level the current level
     */
    public void process(long ticks, Level level) {
        if (isModified()) {
            SimpleVector pos = getBackPosition();
            getTranslationMatrix().setIdentity();
            translate(pos);
            setModified(false);
        } else {
            // Interpolation der Translation
            SimpleVector spd = new SimpleVector(getBackSpeed());
            spd.scalarMul(ticks);
            translate(spd);
        }
        ParticleManager parMan = level.getParticleManager();

        if (parMan != null) {
            if (Math.random() < 0.5f * ticks) {
                SimpleVector vel = getBackSpeed().normalize();
                vel.x += 0.05 - Math.random() / 10f;
                vel.y += 0.05 - Math.random() / 10f;
                vel.z += 0.05 - Math.random() / 10f;
                vel.scalarMul( -1.1f);
                parMan.addParticle(getBackPosition(), vel, 1000L, "particle");
            }
        }
    }
}
