package feud.testgame.view;

import com.threed.jpct.*;
import feud.testgame.model.*;

/**
 * A client object is the visual presentation of a LocalObject. Because we are using a 3D presentation
 * here, it simply an extended version of jPCT's Object3D class. In addition to a normal Object3D, a ClientObject
 * holds a few "backbuffer" values, that has been set based on new data from the server but that hasn't been
 * used in the view until now.
 */
public class ClientObject extends Object3D {

    private boolean isModified = false;
    private Matrix backRot = new Matrix();
    private SimpleVector backPos = new SimpleVector();
    private SimpleVector speed = new SimpleVector();
    private int backAnim = 0;
    private int backAnimSpeed = 0;
    private long lastAccess = 0;
    private long ticks = 0;
    private int lastSequence = -1;
    private int lastAnimSpeed = 1;
    private boolean transfered = false;
    private Object3D child = null;

    /**
     * Creates a new, empty client object. This can be used as a "glue" object between other, real, objects.
     * Just like the dummy Object3D in jPCT would.
     */
    public ClientObject() {
        super(Object3D.createDummyObj());
        touch();
    }

    /**
     * Create a new client object based on an existing Object3D.
     * @param obj the base object
     */
    public ClientObject(Object3D obj) {
        super(obj);
        touch();
    }

    /**
     * Creates a new client object plus a child object. This is a convience method for handling weapons
     * in this "game". The fact that the actual object is composed of parent and child is reflected in
     * some methods, not in all. Basically, the behaviour of the composed object is similar to an Object3D
     * in jPCT that has a child object attached to it.
     * @param obj the base object
     * @param child the base object of the child
     */
    public ClientObject(Object3D obj, Object3D child) {
        super(obj);
        child = new Object3D(child);
        addChild(child);
        this.child = child;
        touch();
    }

    /**
     * In this default implementation, the method does nothing. You may fill it for a client object
     * if you want to process something that is based on a local object that has been created by this client and
     * that should be treated different from what process() itself does.
     * @param ticks time passed since the last run
     * @param level the current level
     * @param lo the LocalObject that this client object represents. Client objects and local objects are not
     * bound directly, which is why the client object doesn't know of its local one.
     */
    public void processLocal(long ticks, Level level, LocalObject lo) {
    }

    /**
     * In this default implementation, the method does nothing. This is for all kinds of processing like playing
     * animation, doing the movement etc...
     * @param ticks time passed since the last run
     * @param level the current level
     */
    public void process(long ticks, Level level) {
    }

    /**
     * Adds the object as well as a child (if any) to the world. This is similar to World.addObject() from the
     * jPCT framework but it takes care of the child object.
     * @param world the world
     */
    public void addToWorld(World world) {
        world.addObject(this);
        if (child != null) {
            world.addObject(child);
        }
    }

    /**
     * Removes the object as well as a child (if any) from the world. This is similar to World.removeObject() from the
     * jPCT framework but it takes care of the child object.
     * @param world the world
     */
    public void removeFromWorld(World world) {
        world.removeObject(this);
        if (child != null) {
            world.removeObject(child);
        }
    }

    public void animate(float ix, int seq) {
        super.animate(ix, seq);
        if (child != null) {
            child.animate(ix, seq);
        }
    }

    public void setVisibility(boolean visi) {
        super.setVisibility(visi);
        if (child != null) {
            child.setVisibility(visi);
        }
    }

    /**
     * A short cut to set the clamping mode for this object's animation (and it child object).
     * @param mode the new clamping mode like in jPCT's Animation class.
     */
    public void setClampingMode(int mode) {
        super.getAnimationSequence().setClampingMode(mode);
        if (child != null) {
            child.getAnimationSequence().setClampingMode(mode);
        }
    }

    /**
     * Determines if the object is "old". In this context, an object is old if it hasn't been
     * touched for more than 5 seconds.
     * @return boolean is it old?
     */
    public boolean isOld() {
        return System.currentTimeMillis() - lastAccess >= 5000;
    }

    /**
     * "Touches" the object, i.e. it marks the object as being used with a timestamp.
     */
    public void touch() {
        lastAccess = System.currentTimeMillis();
    }

    /**
     * Marks this object as being transfered from the server (the model of course, not this object directly).
     * The purpose is to be able to tell if an object has been initially created on another client or on this one.
     * @param flag true, if it's a remote object
     */
    public void setRemoteFlag(boolean flag) {
        transfered = flag;
    }

    /**
     * Returns if this object is a remote one or a local one.
     * @return boolean is it remote?
     */
    public boolean isRemote() {
        return transfered;
    }

    /**
     * Marks this object has being modified. Modified means, that the server has updated the local object which
     * this object represents.
     * @param mod true, if it has been modified
     */
    public void setModified(boolean mod) {
        isModified = mod;
    }

    /**
     * Returns if this object has been modified. Modified means, that the server has updated the local object which
     * this object represents.
     * @return boolean is it modified?
     */
    public boolean isModified() {
        return isModified;
    }

    /**
     * Sets the back rotation matrix. This matrix has been changed due to a change in the server's data but that change
     * hasn't been reflected in the view yet.
     * @param mat the new rotation matrix
     */
    public void setBackRotationMatrix(Matrix mat) {
        backRot = mat;
        setModified(true);
    }

    /**
     * Returns the back rotation matrix.
     * @return Matrix the matrix
     */
    public Matrix getBackRotationMatrix() {
        return backRot;
    }

    /**
     * Sets the back position. This position has been changed due to a change in the server's data but that change
     * hasn't been reflected in the view yet.
     * @param pos the new position
     */
    public void setBackPosition(SimpleVector pos) {
        backPos = pos;
        setModified(true);
    }

    /**
     * Returns the back position.
     * @return SimpleVector the position
     */
    public SimpleVector getBackPosition() {
        return backPos;
    }

    /**
     * Sets the back speed vector. This speed vector has been changed due to a change in the server's data but that change
     * hasn't been reflected in the view yet.
     * @param spd the new speed vector
     */
    public void setBackSpeed(SimpleVector spd) {
        speed = spd;
        setModified(true);
    }

    /**
     * Returns the nback speed vector.
     * @return SimpleVector the vector
     */
    public SimpleVector getBackSpeed() {
        return speed;
    }

    /**
     * Updates the backbuffer of this object with data from a LocalObject. Usually, this object has been
     * transfered from the server and contains a new position, speed etc...
     * @param lo the local object that contains the new data
     */
    public void setToLocalObject(LocalObject lo) {
        setBackRotationMatrix(lo.getRotation());
        setBackPosition(lo.getPosition());
        setBackSpeed(lo.getSpeed());
        setBackAnimation(lo.getAnimation());
        setBackAnimationSpeed(lo.getAnimationSpeed());
    }

    /**
     * Resets the animation ticks. A client object doesn't necessarily use an animation, but the LocalObject
     * itself contains the information. Therefore, the basic animation methods are within this class, even if
     * it may not use them.
     */
    public void resetTicks() {
        ticks = 0;
    }

    /**
     * Gets the animation ticks passed.
     * @return long the ticks
     */
    public long getTicks() {
        return ticks;
    }

    /**
     * Adds a number of ticks to the current tick count.
     * @param val the number of ticks to add
     */
    public void addToTicks(long val) {
        ticks += val;
        if (ticks < 0) {
            ticks = lastAnimSpeed - (( -ticks) % lastAnimSpeed);
        }
    }

    /**
     * Sets the last animation sequence. This is the one that has been used in the last displayed frame.
     * @param seq the sequence
     */
    public void setLastSequence(int seq) {
        lastSequence = seq;
    }

    /**
     * Returns the last animation sequence. This is the one that has been used in the last displayed frame.
     * @return int the sequence
     */
    public int getLastSequence() {
        return lastSequence;
    }

    /**
     * Sets the animation speed of the currently displayed animation.
     * @param speed the speed
     */
    public void setAnimSpeed(int speed) {
        lastAnimSpeed = speed;
    }

    /**
     * Sets the animtion speed of the backbuffer animation.
     * @param speed the speed
     */
    public void setBackAnimationSpeed(int speed) {
        backAnimSpeed = speed;
    }

    /**
     * Gets the speed of the backbuffer animation.
     * @return int the speed
     */
    public int getBackAnimationSpeed() {
        return backAnimSpeed;
    }

    /**
     * Sets the backbuffer animation sequence.
     * @param anim the sequence number
     */
    public void setBackAnimation(int anim) {
        backAnim = anim;
    }

    /**
     * Returns the backbuffer animation sequence.
     * @return int the sequence number
     */
    public int getBackAnimation() {
        return backAnim;
    }
}
