package feud.testgame.view;

import java.util.*;
import com.threed.jpct.*;

import feud.testgame.model.*;

/**
 * A manager class for handling client objects. This manager holds references to all client objects that this client
 * knows of (event remote ones).
 */
public class ClientObjectManager {

    private Map objects = null;
    private ClientObjectFactory cFac = ClientObjectFactory.getInstance();

    /**
     * Creates a new manager.
     */
    public ClientObjectManager() {
        objects = new HashMap();
    }

    /**
     * Returns all client objects that this manager knows of.
     * @return Collection the client objects
     */
    public Collection getClientObjects() {
        return objects.values();
    }

    /**
     * Returns the underlying data model of a view object.
     * @param co the view object
     * @return LocalObject the underlying local object that the view object represents
     */
    public LocalObject getLocalObjectToView(ClientObject co) {
        for (Iterator itty = objects.entrySet().iterator(); itty.hasNext(); ) {
            Map.Entry entry = (Map.Entry) itty.next();
            if (co.equals(entry.getValue())) {
                return (LocalObject) entry.getKey();
            }
        }
        return null;
    }

    /**
     * Gets a client object that has been created on a remote client with a given id. If no such
     * object exists, it returns null.
     * @param objID the id of the object
     * @param clientID the id of the client
     * @return ClientObject the object or null
     */
    public ClientObject getClientObject(int objID, int clientID) {
        LocalObject dummy = new LocalObject(clientID);
        dummy.setObjectID(objID);
        if (objects.containsKey(dummy)) {
            return (ClientObject) objects.get(dummy);
        }
        return null;
    }


    /**
     * Returns the view object for a given local object. If the manager already knows of this object, it returns
     * that one. If it doesn't, it creates a new one and returns that.
     * @param lo the local object on which the view is based
     * @param world the current world (needed, because the manager has to add the new view object to it)
     * @param bulMan the manager for the local bullets (because collisions of local bullets with this object has be determined)
     * @return ClientObject the object
     */
    public ClientObject getOrCreateClientObject(LocalObject lo, World world, LocalBulletManager bulMan) {
        if (objects.containsKey(lo)) {
            return (ClientObject) objects.get(lo);
        }
        ClientObject co = cFac.create(lo.getType());
        objects.put(lo, co);
        co.addToWorld(world);
        co.addCollisionListener(new ClientObjectCollisionListener(bulMan, lo));
        return co;
    }

    /**
     * Removes all objects from the manager and the world.
     * @param world the current world
     */
    public void removeAll(World world) {
        List tmp = new ArrayList(objects.values());
        for (Iterator itty = tmp.iterator(); itty.hasNext(); ) {
            ClientObject co = (ClientObject) itty.next();
            remove(co, world);
        }
    }

    /**
     * Removes a single object from the manager and the world.
     * @param co the object to remove
     * @param world the world
     */
    public void remove(ClientObject co, World world) {
        co.removeFromWorld(world);
        Set entries = objects.entrySet();
        for (Iterator itty = entries.iterator(); itty.hasNext(); ) {
            Map.Entry es = (Map.Entry) itty.next();
            if (es.getValue().equals(co)) {
                itty.remove();
            }
        }
    }
}
